@@ -23,12 +23,11 @@ module Hledger.Data.Journal (
   filterJournalPostings,
   filterJournalTransactions,
   -- * Querying
-  journalAccountInfo,
   journalAccountNames,
   journalAccountNamesUsed,
-  journalAmountAndPriceCommodities,
+  -- journalAmountAndPriceCommodities,
   journalAmounts,
-  journalCanonicalCommodities,
+  -- journalCanonicalCommodities,
   journalDateSpan,
   journalFilePath,
   journalFilePaths,
@@ -43,7 +42,6 @@ module Hledger.Data.Journal (
   journalEquityAccountQuery,
   journalCashAccountQuery,
   -- * Misc
-  groupPostings,
   matchpats,
   nullctx,
   nulljournal,
@@ -53,7 +51,7 @@ module Hledger.Data.Journal (
 )
 where
 import Data.List
-import Data.Map (findWithDefault, (!), toAscList)
+-- import Data.Map (findWithDefault)
 import Data.Ord
 import Data.Time.Calendar
 import Data.Time.LocalTime
@@ -62,14 +60,13 @@ import Safe (headDef)
 import System.Time (ClockTime(TOD))
 import Test.HUnit
 import Text.Printf
-import qualified Data.Map as Map
+import qualified Data.Map as M
 
 import Hledger.Utils
 import Hledger.Data.Types
 import Hledger.Data.AccountName
-import Hledger.Data.Account()
 import Hledger.Data.Amount
-import Hledger.Data.Commodity
+-- import Hledger.Data.Commodity
 import Hledger.Data.Dates
 import Hledger.Data.Transaction
 import Hledger.Data.Posting
@@ -78,13 +75,14 @@ import Hledger.Query
 
 
 instance Show Journal where
-    show j = printf "Journal %s with %d transactions, %d accounts: %s"
+    show j = printf "Journal %s with %d transactions, %d accounts: %s, commodity styles: %s"
              (journalFilePath j)
              (length (jtxns j) +
               length (jmodifiertxns j) +
               length (jperiodictxns j))
              (length accounts)
              (show accounts)
+             (show $ jcommoditystyles j)
              -- ++ (show $ journalTransactions l)
              where accounts = flatten $ journalAccountNameTree j
 
@@ -110,10 +108,11 @@ nulljournal = Journal { jmodifiertxns = []
                       , jContext = nullctx
                       , files = []
                       , filereadtime = TOD 0 0
+                      , jcommoditystyles = M.fromList []
                       }
 
 nullctx :: JournalContext
-nullctx = Ctx { ctxYear = Nothing, ctxCommodity = Nothing, ctxAccount = [], ctxAliases = [] }
+nullctx = Ctx { ctxYear = Nothing, ctxCommodityAndStyle = Nothing, ctxAccount = [], ctxAliases = [] }
 
 journalFilePath :: Journal -> FilePath
 journalFilePath = fst . mainfile
@@ -372,24 +371,28 @@ journalFinalise tclock tlocal path txt ctx j@Journal{files=fs} =
 -- amounts and working out the canonical commodities, since balancing
 -- depends on display precision. Reports only the first error encountered.
 journalBalanceTransactions :: Journal -> Either String Journal
-journalBalanceTransactions j@Journal{jtxns=ts} =
+journalBalanceTransactions j@Journal{jtxns=ts, jcommoditystyles=ss} =
   case sequence $ map balance ts of Right ts' -> Right j{jtxns=ts'}
                                     Left e    -> Left e
-      where balance = balanceTransaction (Just $ journalCanonicalCommodities j)
+      where balance = balanceTransaction (Just ss)
 
 -- | Convert all the journal's posting amounts (not price amounts) to
 -- their canonical display settings. Ie, all amounts in a given
 -- commodity will use (a) the display settings of the first, and (b)
 -- the greatest precision, of the posting amounts in that commodity.
 journalCanonicaliseAmounts :: Journal -> Journal
-journalCanonicaliseAmounts j@Journal{jtxns=ts} = j{jtxns=map fixtransaction ts}
+journalCanonicaliseAmounts j@Journal{jtxns=ts} = j''
     where
+      j'' = j'{jtxns=map fixtransaction ts}
+      j' = j{jcommoditystyles = canonicalStyles $ journalAmounts j}
       fixtransaction t@Transaction{tpostings=ps} = t{tpostings=map fixposting ps}
       fixposting p@Posting{pamount=a} = p{pamount=fixmixedamount a}
       fixmixedamount (Mixed as) = Mixed $ map fixamount as
-      fixamount a@Amount{commodity=c} = a{commodity=fixcommodity c}
-      fixcommodity c@Commodity{symbol=s} = findWithDefault c s canonicalcommoditymap
-      canonicalcommoditymap = journalCanonicalCommodities j
+      fixamount a@Amount{acommodity=c} = a{astyle=journalCommodityStyle j' c}
+
+-- | Get this journal's canonical amount style for the given commodity, or the null style.
+journalCommodityStyle :: Journal -> Commodity -> AmountStyle
+journalCommodityStyle j c = M.findWithDefault amountstyle c $ jcommoditystyles j
 
 -- -- | Apply this journal's historical price records to unpriced amounts where possible.
 -- journalApplyHistoricalPrices :: Journal -> Journal
@@ -424,30 +427,34 @@ journalConvertAmountsToCost j@Journal{jtxns=ts} = j{jtxns=map fixtransaction ts}
       fixtransaction t@Transaction{tpostings=ps} = t{tpostings=map fixposting ps}
       fixposting p@Posting{pamount=a} = p{pamount=fixmixedamount a}
       fixmixedamount (Mixed as) = Mixed $ map fixamount as
-      fixamount = canonicaliseAmountCommodity (Just $ journalCanonicalCommodities j) . costOfAmount
+      fixamount = canonicaliseAmount (jcommoditystyles j) . costOfAmount
 
--- | Get this journal's unique, display-preference-canonicalised commodities, by symbol.
-journalCanonicalCommodities :: Journal -> Map.Map String Commodity
-journalCanonicalCommodities j = canonicaliseCommodities $ journalAmountCommodities j
+-- -- | Get this journal's unique, display-preference-canonicalised commodities, by symbol.
+-- journalCanonicalCommodities :: Journal -> M.Map String Commodity
+-- journalCanonicalCommodities j = canonicaliseCommodities $ journalAmountCommodities j
 
--- | Get all this journal's amounts' commodities, in the order parsed.
-journalAmountCommodities :: Journal -> [Commodity]
-journalAmountCommodities = map commodity . concatMap amounts . journalAmounts
+-- -- | Get all this journal's amounts' commodities, in the order parsed.
+-- journalAmountCommodities :: Journal -> [Commodity]
+-- journalAmountCommodities = map acommodity . concatMap amounts . journalAmounts
 
--- | Get all this journal's amount and price commodities, in the order parsed.
-journalAmountAndPriceCommodities :: Journal -> [Commodity]
-journalAmountAndPriceCommodities = concatMap amountCommodities . concatMap amounts . journalAmounts
+-- -- | Get all this journal's amount and price commodities, in the order parsed.
+-- journalAmountAndPriceCommodities :: Journal -> [Commodity]
+-- journalAmountAndPriceCommodities = concatMap amountCommodities . concatMap amounts . journalAmounts
 
--- | Get this amount's commodity and any commodities referenced in its price.
-amountCommodities :: Amount -> [Commodity]
-amountCommodities Amount{commodity=c,price=p} =
-    case p of Nothing -> [c]
-              Just (UnitPrice ma)  -> c:(concatMap amountCommodities $ amounts ma)
-              Just (TotalPrice ma) -> c:(concatMap amountCommodities $ amounts ma)
+-- -- | Get this amount's commodity and any commodities referenced in its price.
+-- amountCommodities :: Amount -> [Commodity]
+-- amountCommodities Amount{acommodity=c,aprice=p} =
+--     case p of Nothing -> [c]
+--               Just (UnitPrice ma)  -> c:(concatMap amountCommodities $ amounts ma)
+--               Just (TotalPrice ma) -> c:(concatMap amountCommodities $ amounts ma)
 
--- | Get all this journal's amounts, in the order parsed.
-journalAmounts :: Journal -> [MixedAmount]
-journalAmounts = map pamount . journalPostings
+-- | Get all this journal's (mixed) amounts, in the order parsed.
+journalMixedAmounts :: Journal -> [MixedAmount]
+journalMixedAmounts = map pamount . journalPostings
+
+-- | Get all this journal's component amounts, roughly in the order parsed.
+journalAmounts :: Journal -> [Amount]
+journalAmounts = concatMap flatten . journalMixedAmounts where flatten (Mixed as) = as
 
 -- | The (fully specified) date span containing this journal's transactions,
 -- or DateSpan Nothing Nothing if there are none.
@@ -477,212 +484,9 @@ isnegativepat = (negateprefix `isPrefixOf`)
 
 abspat pat = if isnegativepat pat then drop (length negateprefix) pat else pat
 
--- | Calculate the account tree and all account balances from a journal's
--- postings, returning the results for efficient lookup.
-journalAccountInfo :: Journal -> (Tree AccountName, Map.Map AccountName Account)
-journalAccountInfo j = (ant, amap)
-    where
-      (ant, psof, _, inclbalof) = (groupPostings . journalPostings) j
-      amap = Map.fromList [(a, acctinfo a) | a <- flatten ant]
-      acctinfo a = Account a (psof a) (inclbalof a)
-
-tests_journalAccountInfo = [
- "journalAccountInfo" ~: do
-   let (t,m) = journalAccountInfo samplejournal
-   assertEqual "account tree"
-    (Node "top" [
-      Node "assets" [
-       Node "assets:bank" [
-        Node "assets:bank:checking" [],
-        Node "assets:bank:saving" []
-        ],
-       Node "assets:cash" []
-       ],
-      Node "expenses" [
-       Node "expenses:food" [],
-       Node "expenses:supplies" []
-       ],
-      Node "income" [
-       Node "income:gifts" [],
-       Node "income:salary" []
-       ],
-      Node "liabilities" [
-       Node "liabilities:debts" []
-       ]
-      ]
-     )
-    t
-   mapM_ 
-         (\(e,a) -> assertEqual "" e a)
-         (zip [
-               ("assets",Account "assets" [] (Mixed [dollars (-1)]))
-              ,("assets:bank",Account "assets:bank" [] (Mixed [dollars 1]))
-              ,("assets:bank:checking",Account "assets:bank:checking" [
-                  Posting {
-                    pstatus=False,
-                    paccount="assets:bank:checking",
-                    pamount=(Mixed [dollars 1]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  },
-                  Posting {
-                    pstatus=False,
-                    paccount="assets:bank:checking",
-                    pamount=(Mixed [dollars 1]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  },
-                  Posting {
-                    pstatus=False,
-                    paccount="assets:bank:checking",
-                    pamount=(Mixed [dollars (-1)]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  },
-                  Posting {
-                    pstatus=False,
-                    paccount="assets:bank:checking",
-                    pamount=(Mixed [dollars (-1)]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  }
-                  ] (Mixed [nullamt]))
-              ,("assets:bank:saving",Account "assets:bank:saving" [
-                  Posting {
-                    pstatus=False,
-                    paccount="assets:bank:saving",
-                    pamount=(Mixed [dollars 1]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  }
-                  ] (Mixed [dollars 1]))
-              ,("assets:cash",Account "assets:cash" [
-                  Posting {
-                    pstatus=False,
-                    paccount="assets:cash",
-                    pamount=(Mixed [dollars (-2)]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  }
-                ] (Mixed [dollars (-2)]))
-              ,("expenses",Account "expenses" [] (Mixed [dollars 2]))
-              ,("expenses:food",Account "expenses:food" [
-                  Posting {
-                    pstatus=False,
-                    paccount="expenses:food",
-                    pamount=(Mixed [dollars 1]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  }
-                ] (Mixed [dollars 1]))
-              ,("expenses:supplies",Account "expenses:supplies" [
-                  Posting {
-                    pstatus=False,
-                    paccount="expenses:supplies",
-                    pamount=(Mixed [dollars 1]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  }
-                ] (Mixed [dollars 1]))
-              ,("income",Account "income" [] (Mixed [dollars (-2)]))
-              ,("income:gifts",Account "income:gifts" [
-                  Posting {
-                    pstatus=False,
-                    paccount="income:gifts",
-                    pamount=(Mixed [dollars (-1)]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  }
-                ] (Mixed [dollars (-1)]))
-              ,("income:salary",Account "income:salary" [
-                  Posting {
-                    pstatus=False,
-                    paccount="income:salary",
-                    pamount=(Mixed [dollars (-1)]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  }
-                  ] (Mixed [dollars (-1)]))
-              ,("liabilities",Account "liabilities" [] (Mixed [dollars 1]))
-              ,("liabilities:debts",Account "liabilities:debts" [
-                  Posting {
-                    pstatus=False,
-                    paccount="liabilities:debts",
-                    pamount=(Mixed [dollars 1]),
-                    pcomment="",
-                    ptype=RegularPosting,
-                    ptags=[],
-                    ptransaction=Nothing
-                  }
-                ] (Mixed [dollars 1]))
-              ,("top",Account "top" [] (Mixed [nullamt]))
-             ]
-             (toAscList m)
-         )
- ]
-
--- | Given a list of postings, return an account name tree and three query
--- functions that fetch postings, subaccount-excluding-balance and
--- subaccount-including-balance by account name.
-groupPostings :: [Posting] -> (Tree AccountName,
-                               (AccountName -> [Posting]),
-                               (AccountName -> MixedAmount),
-                               (AccountName -> MixedAmount))
-groupPostings ps = (ant, psof, exclbalof, inclbalof)
-    where
-      anames = sort $ nub $ map paccount ps
-      ant = accountNameTreeFrom $ expandAccountNames anames
-      allanames = flatten ant
-      pmap = Map.union (postingsByAccount ps) (Map.fromList [(a,[]) | a <- allanames])
-      psof = (pmap !)
-      balmap = Map.fromList $ flatten $ calculateBalances ant psof
-      exclbalof = fst . (balmap !)
-      inclbalof = snd . (balmap !)
-
--- | Add subaccount-excluding and subaccount-including balances to a tree
--- of account names somewhat efficiently, given a function that looks up
--- transactions by account name.
-calculateBalances :: Tree AccountName -> (AccountName -> [Posting]) -> Tree (AccountName, (MixedAmount, MixedAmount))
-calculateBalances ant psof = addbalances ant
-    where
-      addbalances (Node a subs) = Node (a,(bal,bal+subsbal)) subs'
-          where
-            bal         = sumPostings $ psof a
-            subsbal     = sum $ map (snd . snd . root) subs'
-            subs'       = map addbalances subs
-
--- | Convert a list of postings to a map from account name to that
--- account's postings.
-postingsByAccount :: [Posting] -> Map.Map AccountName [Posting]
-postingsByAccount ps = m'
-    where
-      sortedps = sortBy (comparing paccount) ps
-      groupedps = groupBy (\p1 p2 -> paccount p1 == paccount p2) sortedps
-      m' = Map.fromList [(paccount $ head g, g) | g <- groupedps]
-
 -- debug helpers
--- traceAmountPrecision a = trace (show $ map (precision . commodity) $ amounts a) a
--- tracePostingsCommodities ps = trace (show $ map ((map (precision . commodity) . amounts) . pamount) ps) ps
+-- traceAmountPrecision a = trace (show $ map (precision . acommodity) $ amounts a) a
+-- tracePostingsCommodities ps = trace (show $ map ((map (precision . acommodity) . amounts) . pamount) ps) ps
 
 -- tests
 
@@ -709,10 +513,9 @@ postingsByAccount ps = m'
 --     liabilities:debts  $1
 --     assets:bank:checking
 --
-Right samplejournal = journalBalanceTransactions $ Journal
-          [] 
-          [] 
-          [
+Right samplejournal = journalBalanceTransactions $ 
+         nulljournal
+         {jtxns = [
            txnTieKnot $ Transaction {
              tdate=parsedate "2008/01/01",
              teffectivedate=Nothing,
@@ -725,7 +528,7 @@ Right samplejournal = journalBalanceTransactions $ Journal
               Posting {
                 pstatus=False,
                 paccount="assets:bank:checking",
-                pamount=(Mixed [dollars 1]),
+                pamount=(Mixed [usd 1]),
                 pcomment="",
                 ptype=RegularPosting,
                 ptags=[],
@@ -756,7 +559,7 @@ Right samplejournal = journalBalanceTransactions $ Journal
               Posting {
                 pstatus=False,
                 paccount="assets:bank:checking",
-                pamount=(Mixed [dollars 1]),
+                pamount=(Mixed [usd 1]),
                 pcomment="",
                 ptype=RegularPosting,
                 ptags=[],
@@ -787,7 +590,7 @@ Right samplejournal = journalBalanceTransactions $ Journal
               Posting {
                 pstatus=False,
                 paccount="assets:bank:saving",
-                pamount=(Mixed [dollars 1]),
+                pamount=(Mixed [usd 1]),
                 pcomment="",
                 ptype=RegularPosting,
                 ptags=[],
@@ -796,7 +599,7 @@ Right samplejournal = journalBalanceTransactions $ Journal
               Posting {
                 pstatus=False,
                 paccount="assets:bank:checking",
-                pamount=(Mixed [dollars (-1)]),
+                pamount=(Mixed [usd (-1)]),
                 pcomment="",
                 ptype=RegularPosting,
                 ptags=[],
@@ -818,7 +621,7 @@ Right samplejournal = journalBalanceTransactions $ Journal
               Posting {
                 pstatus=False,
                 paccount="expenses:food",
-                pamount=(Mixed [dollars 1]),
+                pamount=(Mixed [usd 1]),
                 pcomment="",
                 ptype=RegularPosting,
                 ptags=[],
@@ -827,7 +630,7 @@ Right samplejournal = journalBalanceTransactions $ Journal
               Posting {
                 pstatus=False,
                 paccount="expenses:supplies",
-                pamount=(Mixed [dollars 1]),
+                pamount=(Mixed [usd 1]),
                 pcomment="",
                 ptype=RegularPosting,
                 ptags=[],
@@ -858,7 +661,7 @@ Right samplejournal = journalBalanceTransactions $ Journal
               Posting {
                 pstatus=False,
                 paccount="liabilities:debts",
-                pamount=(Mixed [dollars 1]),
+                pamount=(Mixed [usd 1]),
                 pcomment="",
                 ptype=RegularPosting,
                 ptags=[],
@@ -867,7 +670,7 @@ Right samplejournal = journalBalanceTransactions $ Journal
               Posting {
                 pstatus=False,
                 paccount="assets:bank:checking",
-                pamount=(Mixed [dollars (-1)]),
+                pamount=(Mixed [usd (-1)]),
                 pcomment="",
                 ptype=RegularPosting,
                 ptags=[],
@@ -877,19 +680,13 @@ Right samplejournal = journalBalanceTransactions $ Journal
              tpreceding_comment_lines=""
            }
           ]
-          []
-          []
-          ""
-          nullctx
-          []
-          (TOD 0 0)
+         }
 
 tests_Hledger_Data_Journal = TestList $
-    tests_journalAccountInfo
-  -- [
+ [
   -- "query standard account types" ~:
   --  do
   --   let j = journal1
   --   journalBalanceSheetAccountNames j `is` ["assets","assets:a","equity","equity:q","equity:q:qq","liabilities","liabilities:l"]
   --   journalProfitAndLossAccountNames j `is` ["expenses","expenses:e","income","income:i"]
- -- ]
+ ]
