@@ -47,7 +47,6 @@ import Hledger.Data.Types
 import Hledger.Data.Dates
 import Hledger.Data.Posting
 import Hledger.Data.Amount
-import Hledger.Data.Commodity
 
 instance Show Transaction where show = showTransactionUnelided
 
@@ -108,7 +107,7 @@ tests_showTransactionUnelided = [
         nullposting{
           pstatus=True,
           paccount="a",
-          pamount=Mixed [dollars 1, hours 2],
+          pamount=Mixed [usd 1, hrs 2],
           pcomment="pcomment1\npcomment2\n",
           ptype=RegularPosting,
           ptags=[("ptag1","val1"),("ptag2","val2")]
@@ -183,7 +182,7 @@ tests_postingAsLines = [
     nullposting{
       pstatus=True,
       paccount="a",
-      pamount=Mixed [dollars 1, hours 2],
+      pamount=Mixed [usd 1, hrs 2],
       pcomment="pcomment1\npcomment2\n",
       ptype=RegularPosting,
       ptags=[("ptag1","val1"),("ptag2","val2")]
@@ -236,14 +235,15 @@ transactionPostingBalances t = (sumPostings $ realPostings t
 -- | Is this transaction balanced ? A balanced transaction's real
 -- (non-virtual) postings sum to 0, and any balanced virtual postings
 -- also sum to 0.
-isTransactionBalanced :: Maybe (Map.Map String Commodity) -> Transaction -> Bool
-isTransactionBalanced canonicalcommoditymap t =
+isTransactionBalanced :: Maybe (Map.Map Commodity AmountStyle) -> Transaction -> Bool
+isTransactionBalanced styles t =
     -- isReallyZeroMixedAmountCost rsum && isReallyZeroMixedAmountCost bvsum
     isZeroMixedAmount rsum' && isZeroMixedAmount bvsum'
     where
       (rsum, _, bvsum) = transactionPostingBalances t
-      rsum'  = canonicaliseMixedAmountCommodity canonicalcommoditymap $ costOfMixedAmount rsum
-      bvsum' = canonicaliseMixedAmountCommodity canonicalcommoditymap $ costOfMixedAmount bvsum
+      rsum'  = canonicalise $ costOfMixedAmount rsum
+      bvsum' = canonicalise $ costOfMixedAmount bvsum
+      canonicalise = maybe id canonicaliseMixedAmount styles
 
 -- | Ensure this transaction is balanced, possibly inferring a missing
 -- amount or conversion price, or return an error message.
@@ -260,11 +260,11 @@ isTransactionBalanced canonicalcommoditymap t =
 -- and the sum of real postings' amounts is exactly two
 -- non-explicitly-priced amounts in different commodities (likewise
 -- for balanced virtual postings).
-balanceTransaction :: Maybe (Map.Map String Commodity) -> Transaction -> Either String Transaction
-balanceTransaction canonicalcommoditymap t@Transaction{tpostings=ps}
+balanceTransaction :: Maybe (Map.Map Commodity AmountStyle) -> Transaction -> Either String Transaction
+balanceTransaction styles t@Transaction{tpostings=ps}
     | length rwithoutamounts > 1 || length bvwithoutamounts > 1
         = Left $ printerr "could not balance this transaction (too many missing amounts)"
-    | not $ isTransactionBalanced canonicalcommoditymap t''' = Left $ printerr $ nonzerobalanceerror t'''
+    | not $ isTransactionBalanced styles t''' = Left $ printerr $ nonzerobalanceerror t'''
     | otherwise = Right t'''
     where
       -- maybe infer missing amounts
@@ -281,53 +281,53 @@ balanceTransaction canonicalcommoditymap t@Transaction{tpostings=ps}
       -- maybe infer conversion prices, for real postings
       rmixedamountsinorder = map pamount $ realPostings t'
       ramountsinorder = concatMap amounts rmixedamountsinorder
-      rcommoditiesinorder  = map commodity ramountsinorder
+      rcommoditiesinorder  = map acommodity ramountsinorder
       rsumamounts  = amounts $ sum rmixedamountsinorder
       -- assumption: the sum of mixed amounts is normalised (one simple amount per commodity)
-      t'' = if length rsumamounts == 2 && all (isNothing.price) rsumamounts && t'==t
+      t'' = if length rsumamounts == 2 && all ((==NoPrice).aprice) rsumamounts && t'==t
              then t'{tpostings=map inferprice ps}
              else t'
           where
             -- assumption: a posting's mixed amount contains one simple amount
-            inferprice p@Posting{pamount=Mixed [a@Amount{commodity=c,price=Nothing}], ptype=RegularPosting}
-                = p{pamount=Mixed [a{price=conversionprice c}]}
+            inferprice p@Posting{pamount=Mixed [a@Amount{acommodity=c,aprice=NoPrice}], ptype=RegularPosting}
+                = p{pamount=Mixed [a{aprice=conversionprice c}]}
                 where
                   conversionprice c | c == unpricedcommodity
                                         -- assign a balancing price. Use @@ for more exact output when possible.
                                         -- invariant: prices should always be positive. Enforced with "abs"
                                         = if length ramountsinunpricedcommodity == 1
-                                           then Just $ TotalPrice $ Mixed [setAmountPrecision maxprecision $ abs $ targetcommodityamount]
-                                           else Just $ UnitPrice $ Mixed [setAmountPrecision maxprecision $ abs $ targetcommodityamount `divideAmount` (quantity unpricedamount)]
-                                    | otherwise = Nothing
+                                           then TotalPrice $ abs targetcommodityamount `withPrecision` maxprecision
+                                           else UnitPrice $ abs (targetcommodityamount `divideAmount` (aquantity unpricedamount)) `withPrecision` maxprecision
+                                    | otherwise = NoPrice
                       where
-                        unpricedcommodity     = head $ filter (`elem` (map commodity rsumamounts)) rcommoditiesinorder
-                        unpricedamount        = head $ filter ((==unpricedcommodity).commodity) rsumamounts
-                        targetcommodityamount = head $ filter ((/=unpricedcommodity).commodity) rsumamounts
-                        ramountsinunpricedcommodity = filter ((==unpricedcommodity).commodity) ramountsinorder
+                        unpricedcommodity     = head $ filter (`elem` (map acommodity rsumamounts)) rcommoditiesinorder
+                        unpricedamount        = head $ filter ((==unpricedcommodity).acommodity) rsumamounts
+                        targetcommodityamount = head $ filter ((/=unpricedcommodity).acommodity) rsumamounts
+                        ramountsinunpricedcommodity = filter ((==unpricedcommodity).acommodity) ramountsinorder
             inferprice p = p
 
       -- maybe infer prices for balanced virtual postings. Just duplicates the above for now.
       bvmixedamountsinorder = map pamount $ balancedVirtualPostings t''
       bvamountsinorder = concatMap amounts bvmixedamountsinorder
-      bvcommoditiesinorder  = map commodity bvamountsinorder
+      bvcommoditiesinorder  = map acommodity bvamountsinorder
       bvsumamounts  = amounts $ sum bvmixedamountsinorder
-      t''' = if length bvsumamounts == 2 && all (isNothing.price) bvsumamounts && t'==t -- XXX could check specifically for bv amount inferring
+      t''' = if length bvsumamounts == 2 && all ((==NoPrice).aprice) bvsumamounts && t'==t -- XXX could check specifically for bv amount inferring
              then t''{tpostings=map inferprice ps}
              else t''
           where
-            inferprice p@Posting{pamount=Mixed [a@Amount{commodity=c,price=Nothing}], ptype=BalancedVirtualPosting}
-                = p{pamount=Mixed [a{price=conversionprice c}]}
+            inferprice p@Posting{pamount=Mixed [a@Amount{acommodity=c,aprice=NoPrice}], ptype=BalancedVirtualPosting}
+                = p{pamount=Mixed [a{aprice=conversionprice c}]}
                 where
                   conversionprice c | c == unpricedcommodity
                                         = if length bvamountsinunpricedcommodity == 1
-                                           then Just $ TotalPrice $ Mixed [setAmountPrecision maxprecision $ abs $ targetcommodityamount]
-                                           else Just $ UnitPrice $ Mixed [setAmountPrecision maxprecision $ abs $ targetcommodityamount `divideAmount` (quantity unpricedamount)]
-                                    | otherwise = Nothing
+                                           then TotalPrice $ abs targetcommodityamount `withPrecision` maxprecision
+                                           else UnitPrice $ abs (targetcommodityamount `divideAmount` (aquantity unpricedamount)) `withPrecision` maxprecision
+                                    | otherwise = NoPrice
                       where
-                        unpricedcommodity     = head $ filter (`elem` (map commodity bvsumamounts)) bvcommoditiesinorder
-                        unpricedamount        = head $ filter ((==unpricedcommodity).commodity) bvsumamounts
-                        targetcommodityamount = head $ filter ((/=unpricedcommodity).commodity) bvsumamounts
-                        bvamountsinunpricedcommodity = filter ((==unpricedcommodity).commodity) bvamountsinorder
+                        unpricedcommodity     = head $ filter (`elem` (map acommodity bvsumamounts)) bvcommoditiesinorder
+                        unpricedamount        = head $ filter ((==unpricedcommodity).acommodity) bvsumamounts
+                        targetcommodityamount = head $ filter ((/=unpricedcommodity).acommodity) bvsumamounts
+                        bvamountsinunpricedcommodity = filter ((==unpricedcommodity).acommodity) bvamountsinorder
             inferprice p = p
 
       printerr s = intercalate "\n" [s, showTransactionUnelided t]
@@ -340,7 +340,7 @@ nonzerobalanceerror t = printf "could not balance this transaction (%s%s%s)" rms
            | otherwise = "real postings are off by " ++ showMixedAmount (costOfMixedAmount rsum)
       bvmsg | isReallyZeroMixedAmountCost bvsum = ""
             | otherwise = "balanced virtual postings are off by " ++ showMixedAmount (costOfMixedAmount bvsum)
-      sep = if not (null rmsg) && not (null bvmsg) then "; " else ""
+      sep = if not (null rmsg) && not (null bvmsg) then "; " else "" :: String
 
 transactionActualDate :: Transaction -> Day
 transactionActualDate = tdate
@@ -376,8 +376,8 @@ tests_Hledger_Data_Transaction = TestList $ concat [
         ,""
         ])
        (let t = Transaction (parsedate "2007/01/28") Nothing False "" "coopportunity" "" []
-                [Posting False "expenses:food:groceries" (Mixed [dollars 47.18]) "" RegularPosting [] (Just t)
-                ,Posting False "assets:checking" (Mixed [dollars (-47.18)]) "" RegularPosting [] (Just t)
+                [Posting False "expenses:food:groceries" (Mixed [usd 47.18]) "" RegularPosting [] (Just t)
+                ,Posting False "assets:checking" (Mixed [usd (-47.18)]) "" RegularPosting [] (Just t)
                 ] ""
         in showTransaction t)
 
@@ -390,8 +390,8 @@ tests_Hledger_Data_Transaction = TestList $ concat [
         ,""
         ])
        (let t = Transaction (parsedate "2007/01/28") Nothing False "" "coopportunity" "" []
-                [Posting False "expenses:food:groceries" (Mixed [dollars 47.18]) "" RegularPosting [] (Just t)
-                ,Posting False "assets:checking" (Mixed [dollars (-47.18)]) "" RegularPosting [] (Just t)
+                [Posting False "expenses:food:groceries" (Mixed [usd 47.18]) "" RegularPosting [] (Just t)
+                ,Posting False "assets:checking" (Mixed [usd (-47.18)]) "" RegularPosting [] (Just t)
                 ] ""
         in showTransactionUnelided t)
 
@@ -406,8 +406,8 @@ tests_Hledger_Data_Transaction = TestList $ concat [
         ])
        (showTransaction
         (txnTieKnot $ Transaction (parsedate "2007/01/28") Nothing False "" "coopportunity" "" []
-         [Posting False "expenses:food:groceries" (Mixed [dollars 47.18]) "" RegularPosting [] Nothing
-         ,Posting False "assets:checking" (Mixed [dollars (-47.19)]) "" RegularPosting [] Nothing
+         [Posting False "expenses:food:groceries" (Mixed [usd 47.18]) "" RegularPosting [] Nothing
+         ,Posting False "assets:checking" (Mixed [usd (-47.19)]) "" RegularPosting [] Nothing
          ] ""))
 
   ,"showTransaction" ~: do
@@ -419,7 +419,7 @@ tests_Hledger_Data_Transaction = TestList $ concat [
         ])
        (showTransaction
         (txnTieKnot $ Transaction (parsedate "2007/01/28") Nothing False "" "coopportunity" "" []
-         [Posting False "expenses:food:groceries" (Mixed [dollars 47.18]) "" RegularPosting [] Nothing
+         [Posting False "expenses:food:groceries" (Mixed [usd 47.18]) "" RegularPosting [] Nothing
          ] ""))
 
   ,"showTransaction" ~: do
@@ -444,7 +444,7 @@ tests_Hledger_Data_Transaction = TestList $ concat [
         ])
        (showTransaction
         (txnTieKnot $ Transaction (parsedate "2010/01/01") Nothing False "" "x" "" []
-         [Posting False "a" (Mixed [Amount unknown 1 (Just $ UnitPrice $ Mixed [Amount dollar{precision=0} 2 Nothing])]) "" RegularPosting [] Nothing
+         [Posting False "a" (Mixed [num 1 `at` (usd 2 `withPrecision` 0)]) "" RegularPosting [] Nothing
          ,Posting False "b" missingmixedamt "" RegularPosting [] Nothing
          ] ""))
 
@@ -452,76 +452,87 @@ tests_Hledger_Data_Transaction = TestList $ concat [
      assertBool "detect unbalanced entry, sign error"
                     (isLeft $ balanceTransaction Nothing
                            (Transaction (parsedate "2007/01/28") Nothing False "" "test" "" []
-                            [Posting False "a" (Mixed [dollars 1]) "" RegularPosting [] Nothing,
-                             Posting False "b" (Mixed [dollars 1]) "" RegularPosting [] Nothing
+                            [Posting False "a" (Mixed [usd 1]) "" RegularPosting [] Nothing,
+                             Posting False "b" (Mixed [usd 1]) "" RegularPosting [] Nothing
                             ] ""))
+
      assertBool "detect unbalanced entry, multiple missing amounts"
                     (isLeft $ balanceTransaction Nothing
                            (Transaction (parsedate "2007/01/28") Nothing False "" "test" "" []
                             [Posting False "a" missingmixedamt "" RegularPosting [] Nothing,
                              Posting False "b" missingmixedamt "" RegularPosting [] Nothing
                             ] ""))
+
      let e = balanceTransaction Nothing (Transaction (parsedate "2007/01/28") Nothing False "" "" "" []
-                           [Posting False "a" (Mixed [dollars 1]) "" RegularPosting [] Nothing,
+                           [Posting False "a" (Mixed [usd 1]) "" RegularPosting [] Nothing,
                             Posting False "b" missingmixedamt "" RegularPosting [] Nothing
                            ] "")
      assertBool "balanceTransaction allows one missing amount" (isRight e)
      assertEqual "balancing amount is inferred"
-                     (Mixed [dollars (-1)])
+                     (Mixed [usd (-1)])
                      (case e of
                         Right e' -> (pamount $ last $ tpostings e')
                         Left _ -> error' "should not happen")
+
      let e = balanceTransaction Nothing (Transaction (parsedate "2011/01/01") Nothing False "" "" "" []
-                           [Posting False "a" (Mixed [dollars 1.35]) "" RegularPosting [] Nothing,
-                            Posting False "b" (Mixed [euros   (-1)]) "" RegularPosting [] Nothing
+                           [Posting False "a" (Mixed [usd 1.35]) "" RegularPosting [] Nothing,
+                            Posting False "b" (Mixed [eur (-1)]) "" RegularPosting [] Nothing
                            ] "")
      assertBool "balanceTransaction can infer conversion price" (isRight e)
      assertEqual "balancing conversion price is inferred"
-                     (Mixed [Amount{commodity=dollar{precision=2},
-                                    quantity=1.35,
-                                    price=(Just $ TotalPrice $ Mixed [Amount{commodity=euro{precision=maxprecision},
-                                                                             quantity=1,
-                                                                             price=Nothing}])}])
+                     (Mixed [usd 1.35 @@ (eur 1 `withPrecision` maxprecision)])
                      (case e of
                         Right e' -> (pamount $ head $ tpostings e')
                         Left _ -> error' "should not happen")
 
+     assertBool "balanceTransaction balances based on cost if there are unit prices" (isRight $
+       balanceTransaction Nothing (Transaction (parsedate "2011/01/01") Nothing False "" "" "" []
+                           [Posting False "a" (Mixed [usd 1 `at` eur 2]) "" RegularPosting [] Nothing
+                           ,Posting False "a" (Mixed [usd (-2) `at` eur 1]) "" RegularPosting [] Nothing
+                           ] ""))
+
+     assertBool "balanceTransaction balances based on cost if there are total prices" (isRight $
+       balanceTransaction Nothing (Transaction (parsedate "2011/01/01") Nothing False "" "" "" []
+                           [Posting False "a" (Mixed [usd 1    @@ eur 1]) "" RegularPosting [] Nothing
+                           ,Posting False "a" (Mixed [usd (-2) @@ eur 1]) "" RegularPosting [] Nothing
+                           ] ""))
+
   ,"isTransactionBalanced" ~: do
      let t = Transaction (parsedate "2009/01/01") Nothing False "" "a" "" []
-             [Posting False "b" (Mixed [dollars 1.00]) "" RegularPosting [] (Just t)
-             ,Posting False "c" (Mixed [dollars (-1.00)]) "" RegularPosting [] (Just t)
+             [Posting False "b" (Mixed [usd 1.00]) "" RegularPosting [] (Just t)
+             ,Posting False "c" (Mixed [usd (-1.00)]) "" RegularPosting [] (Just t)
              ] ""
      assertBool "detect balanced" (isTransactionBalanced Nothing t)
      let t = Transaction (parsedate "2009/01/01") Nothing False "" "a" "" []
-             [Posting False "b" (Mixed [dollars 1.00]) "" RegularPosting [] (Just t)
-             ,Posting False "c" (Mixed [dollars (-1.01)]) "" RegularPosting [] (Just t)
+             [Posting False "b" (Mixed [usd 1.00]) "" RegularPosting [] (Just t)
+             ,Posting False "c" (Mixed [usd (-1.01)]) "" RegularPosting [] (Just t)
              ] ""
      assertBool "detect unbalanced" (not $ isTransactionBalanced Nothing t)
      let t = Transaction (parsedate "2009/01/01") Nothing False "" "a" "" []
-             [Posting False "b" (Mixed [dollars 1.00]) "" RegularPosting [] (Just t)
+             [Posting False "b" (Mixed [usd 1.00]) "" RegularPosting [] (Just t)
              ] ""
      assertBool "detect unbalanced, one posting" (not $ isTransactionBalanced Nothing t)
      let t = Transaction (parsedate "2009/01/01") Nothing False "" "a" "" []
-             [Posting False "b" (Mixed [dollars 0]) "" RegularPosting [] (Just t)
+             [Posting False "b" (Mixed [usd 0]) "" RegularPosting [] (Just t)
              ] ""
      assertBool "one zero posting is considered balanced for now" (isTransactionBalanced Nothing t)
      let t = Transaction (parsedate "2009/01/01") Nothing False "" "a" "" []
-             [Posting False "b" (Mixed [dollars 1.00]) "" RegularPosting [] (Just t)
-             ,Posting False "c" (Mixed [dollars (-1.00)]) "" RegularPosting [] (Just t)
-             ,Posting False "d" (Mixed [dollars 100]) "" VirtualPosting [] (Just t)
+             [Posting False "b" (Mixed [usd 1.00]) "" RegularPosting [] (Just t)
+             ,Posting False "c" (Mixed [usd (-1.00)]) "" RegularPosting [] (Just t)
+             ,Posting False "d" (Mixed [usd 100]) "" VirtualPosting [] (Just t)
              ] ""
      assertBool "virtual postings don't need to balance" (isTransactionBalanced Nothing t)
      let t = Transaction (parsedate "2009/01/01") Nothing False "" "a" "" []
-             [Posting False "b" (Mixed [dollars 1.00]) "" RegularPosting [] (Just t)
-             ,Posting False "c" (Mixed [dollars (-1.00)]) "" RegularPosting [] (Just t)
-             ,Posting False "d" (Mixed [dollars 100]) "" BalancedVirtualPosting [] (Just t)
+             [Posting False "b" (Mixed [usd 1.00]) "" RegularPosting [] (Just t)
+             ,Posting False "c" (Mixed [usd (-1.00)]) "" RegularPosting [] (Just t)
+             ,Posting False "d" (Mixed [usd 100]) "" BalancedVirtualPosting [] (Just t)
              ] ""
      assertBool "balanced virtual postings need to balance among themselves" (not $ isTransactionBalanced Nothing t)
      let t = Transaction (parsedate "2009/01/01") Nothing False "" "a" "" []
-             [Posting False "b" (Mixed [dollars 1.00]) "" RegularPosting [] (Just t)
-             ,Posting False "c" (Mixed [dollars (-1.00)]) "" RegularPosting [] (Just t)
-             ,Posting False "d" (Mixed [dollars 100]) "" BalancedVirtualPosting [] (Just t)
-             ,Posting False "e" (Mixed [dollars (-100)]) "" BalancedVirtualPosting [] (Just t)
+             [Posting False "b" (Mixed [usd 1.00]) "" RegularPosting [] (Just t)
+             ,Posting False "c" (Mixed [usd (-1.00)]) "" RegularPosting [] (Just t)
+             ,Posting False "d" (Mixed [usd 100]) "" BalancedVirtualPosting [] (Just t)
+             ,Posting False "e" (Mixed [usd (-100)]) "" BalancedVirtualPosting [] (Just t)
              ] ""
      assertBool "balanced virtual postings need to balance among themselves (2)" (isTransactionBalanced Nothing t)
 
