@@ -19,8 +19,9 @@ module Hledger.Read (
        ensureJournalFileExists,
        -- * Parsers used elsewhere
        accountname,
-       amount,
-       amount',
+       amountp,
+       amountp',
+       mamountp',
        -- * Tests
        samplejournal,
        tests_Hledger_Read,
