@@ -61,7 +61,7 @@ import Prelude hiding (getContents)
 import Hledger.Utils.UTF8IOCompat (getContents)
 import Hledger.Utils
 import Hledger.Data.FormatStrings as FormatStrings
-import Hledger.Read.JournalReader (accountname, amount)
+import Hledger.Read.JournalReader (accountname, amountp)
 
 
 reader :: Reader
@@ -98,7 +98,8 @@ nullrules = CsvRules {
       account2Field=Nothing,
       effectiveDateField=Nothing,
       baseAccount="unknown",
-      accountRules=[]
+      accountRules=[],
+      skipLines=0
 }
 
 type CsvRecord = [String]
@@ -137,9 +138,10 @@ readJournalFromCsv mrulesfile csvfile csvdata =
   rules <- liftM (either (throw.userError.show) id) $ parseCsvRulesFile rulesfile
 
   let requiredfields = (maxFieldIndex rules + 1)
-      badrecords = take 1 $ filter ((< requiredfields).length) records
+      realrecords = drop (skipLines rules) records
+      badrecords = take 1 $ filter ((< requiredfields).length) realrecords
   return $ case badrecords of
-            []    -> Right nulljournal{jtxns=sortBy (comparing tdate) $ map (transactionFromCsvRecord rules) records}
+            []    -> Right nulljournal{jtxns=sortBy (comparing tdate) $ map (transactionFromCsvRecord rules) realrecords}
             (_:_) -> Left $ "Parse error: at least one CSV record does not contain a field referenced by the conversion rules file:\n"++(show $ head badrecords)
 
 -- | Ensure there is a conversion rules file at the given path, creating a
@@ -189,6 +191,7 @@ newRulesFileContent = let prognameandversion = "hledger" in
     "# Add rules to this file for more accurate conversion, see\n"++
     "# http://hledger.org/MANUAL.html#convert\n" ++
     "\n" ++
+    "skip-lines 0\n" ++
     "base-account assets:bank:checking\n" ++
     "date-field 0\n" ++
     "description-field 4\n" ++
@@ -251,6 +254,7 @@ definitions = do
    ,effectivedatefield
    ,basecurrency
    ,baseaccount
+   ,skiplines
    ,commentline
    ] <?> "definition"
   return ()
@@ -350,6 +354,12 @@ baseaccount = do
   optional newline
   updateState (\r -> r{baseAccount=v})
 
+skiplines = do
+  string "skip-lines"
+  many1 spacenonewline
+  v <- restofline
+  updateState (\r -> r{skipLines=read v})
+
 accountrule :: GenParser Char CsvRules AccountRule
 accountrule = do
   many blankorcommentline
@@ -426,8 +436,8 @@ transactionFromCsvRecord rules fields =
                                              strnegate s = '-':s
       currency = maybe (fromMaybe "" $ baseCurrency rules) (atDef "" fields) (currencyField rules)
       amountstr'' = currency ++ amountstr'
-      amountparse = runParser amount nullctx "" amountstr''
-      a = either (const nullmixedamt) id amountparse
+      amountparse = runParser amountp nullctx "" amountstr''
+      a = either (const nullmixedamt) mixed amountparse
       -- Using costOfMixedAmount here to allow complex costs like "10 GBP @@ 15 USD".
       -- Aim is to have "10 GBP @@ 15 USD" applied to account "acct", but have "-15USD" applied to "baseacct"
       baseamount = costOfMixedAmount a
