@@ -31,6 +31,7 @@ import Control.Monad.Error (MonadIO)
 import Control.Monad.IO.Class (liftIO)
 import Data.Char
 import Data.List
+import qualified Data.Map as M
 import Data.Maybe
 import Data.Time.Clock
 import Data.Time.LocalTime
@@ -54,7 +55,7 @@ lowercase = map toLower
 uppercase = map toUpper
 
 strip = lstrip . rstrip
-lstrip = dropWhile (`elem` " \t")
+lstrip = dropWhile (`elem` " \t") :: String -> String
 rstrip = reverse . lstrip . reverse
 
 elideLeft width s =
@@ -237,6 +238,8 @@ splitAtElement e l =
 
 -- trees
 
+-- standard tree helpers
+
 root = rootLabel
 subs = subForest
 branches = subForest
@@ -291,16 +294,41 @@ showtree = unlines . filter (regexMatches "[^ \\|]") . lines . drawTree . treema
 showforest :: Show a => Forest a -> String
 showforest = concatMap showtree
 
+
+-- | An efficient-to-build tree suggested by Cale Gibbard, probably
+-- better than accountNameTreeFrom.
+newtype FastTree a = T (M.Map a (FastTree a))
+  deriving (Show, Eq, Ord)
+
+emptyTree = T M.empty
+
+mergeTrees :: (Ord a) => FastTree a -> FastTree a -> FastTree a
+mergeTrees (T m) (T m') = T (M.unionWith mergeTrees m m')
+
+treeFromPath :: [a] -> FastTree a
+treeFromPath []     = T M.empty
+treeFromPath (x:xs) = T (M.singleton x (treeFromPath xs))
+
+treeFromPaths :: (Ord a) => [[a]] -> FastTree a
+treeFromPaths = foldl' mergeTrees emptyTree . map treeFromPath
+
+
 -- debugging
 
+-- more:
+-- http://hackage.haskell.org/packages/archive/TraceUtils/0.1.0.2/doc/html/Debug-TraceUtils.html
+-- http://hackage.haskell.org/packages/archive/trace-call/0.1/doc/html/Debug-TraceCall.html
+-- http://hackage.haskell.org/packages/archive/htrace/0.1/doc/html/Debug-HTrace.html
+-- http://hackage.haskell.org/packages/archive/traced/2009.7.20/doc/html/Debug-Traced.html
+
 -- | trace (print on stdout at runtime) a showable expression
 -- (for easily tracing in the middle of a complex expression)
 strace :: Show a => a -> a
 strace a = trace (show a) a
 
--- | labelled trace - like strace, with a label prepended
-ltrace :: Show a => String -> a -> a
-ltrace l a = trace (l ++ ": " ++ show a) a
+-- | labelled trace showable - like strace, with a label prepended
+lstrace :: Show a => String -> a -> a
+lstrace l a = trace (l ++ ": " ++ show a) a
 
 -- | monadic trace - like strace, but works as a standalone line in a monad
 mtrace :: (Monad m, Show a) => a -> m a
